# distutils: language=c++
#  =============================================================================
#  OVR C API (OVR_CAPI.h) Cython Declaration File 
#  =============================================================================
#
#  ovr_capi.pxd
#
#  Copyright 2018 Matthew Cutone <cutonem(a)yorku.ca> and Laurie M. Wilcox
#  <lmwilcox(a)yorku.ca>; The Centre For Vision Research, York University, 
#  Toronto, Canada
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in 
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# 
"""This file exposes Oculus Rift(TM) C API types and functions, allowing Cython 
extensions to access them. The declarations in the file are contemporaneous
with version 1.24 (retrieved 04.15.2018) of the Oculus Rift(TM) PC SDK. 

"""
from ovr_errorcode cimport ovrResult, ovrErrorInfo
from libc.stdint cimport uintptr_t, uint32_t, int32_t

cdef extern from "OVR_CAPI.h":
    ctypedef char ovrBool
    cdef ovrBool ovrFalse = 0
    cdef ovrBool ovrTrue = 1

    ctypedef struct ovrColorf:
        float r
        float g
        float b
        float a

    ctypedef struct ovrVector2i:
        int x
        int y

    ctypedef struct ovrSizei:
        int w 
        int h

    ctypedef struct ovrRecti:
        ovrVector2i Pos
        ovrSizei Size

    ctypedef struct ovrQuatf:
        float x
        float y
        float z 
        float w

    ctypedef struct ovrVector2f:
        float x
        float y

    ctypedef struct ovrVector3f:
        float x 
        float y 
        float z

    ctypedef struct ovrMatrix4f:
        float[4][4] M

    ctypedef struct ovrPosef:
        ovrQuatf Orientation
        ovrVector3f Position

    ctypedef struct ovrPoseStatef:
        ovrPosef ThePose
        ovrVector3f AngularVelocity 
        ovrVector3f LinearVelocity 
        ovrVector3f AngularAcceleration
        ovrVector3f LinearAcceleration 
        double TimeInSeconds 

    ctypedef struct ovrFovPort:
        float UpTan 
        float DownTan 
        float LeftTan 
        float RightTan 

    ctypedef enum ovrHmdType:
        ovrHmd_None = 0,
        ovrHmd_DK1 = 3,
        ovrHmd_DKHD = 4,
        ovrHmd_DK2 = 6,
        ovrHmd_CB = 8,
        ovrHmd_Other = 9,
        ovrHmd_E3_2015 = 10,
        ovrHmd_ES06 = 11,
        ovrHmd_ES09 = 12,
        ovrHmd_ES11 = 13,
        ovrHmd_CV1 = 14

    ctypedef enum ovrHmdCaps:
        ovrHmdCap_DebugDevice = 0x0010

    ctypedef enum ovrTrackingCaps:
        ovrTrackingCap_Orientation = 0x0010,
        ovrTrackingCap_MagYawCorrection = 0x0020,
        ovrTrackingCap_Position = 0x0040

    ctypedef enum ovrExtensions:
        ovrExtension_TextureLayout_Octilinear = 0 
        
    ctypedef enum ovrEyeType:
        ovrEye_Left = 0, 
        ovrEye_Right = 1,
        ovrEye_Count = 2

    ctypedef enum ovrTrackingOrigin:
        ovrTrackingOrigin_EyeLevel = 0,
        ovrTrackingOrigin_FloorLevel = 1,
        ovrTrackingOrigin_Count = 2 

    ctypedef struct ovrGraphicsLuid:
        char[8] Reserved

    ctypedef struct ovrHmdDesc:
        ovrHmdType Type
        char[64] ProductName
        char[64] Manufacturer
        short VendorId 
        short ProductId 
        char[24] SerialNumber
        short FirmwareMajor
        short FirmwareMinor 
        unsigned int AvailableHmdCaps 
        unsigned int DefaultHmdCaps 
        unsigned int AvailableTrackingCaps
        unsigned int DefaultTrackingCaps 
        ovrFovPort[2] DefaultEyeFov
        ovrFovPort[2] MaxEyeFov
        ovrSizei Resolution 
        float DisplayRefreshRate 

    ctypedef struct ovrHmdStruct
    ctypedef ovrHmdStruct* ovrSession
    ctypedef uint32_t ovrProcessId

    ctypedef enum ovrStatusBits:
        ovrStatus_OrientationTracked = 0x0001, 
        ovrStatus_PositionTracked = 0x0002

    ctypedef struct ovrTrackerDesc:
        float FrustumHFovInRadians 
        float FrustumVFovInRadians 
        float FrustumNearZInMeters
        float FrustumFarZInMeters 

    ctypedef enum ovrTrackerFlags:
        ovrTracker_Connected = 0x0020,
        ovrTracker_PoseTracked = 0x0004

    ctypedef struct ovrTrackerPose:
        unsigned int TrackerFlags
        ovrPosef Pose
        ovrPosef LeveledPose

    ctypedef struct ovrTrackingState:
        ovrPoseStatef HeadPose
        unsigned int StatusFlags
        ovrPoseStatef[2] HandPoses
        unsigned int[2] HandStatusFlags
        ovrPosef CalibratedOrigin

    ctypedef struct ovrEyeRenderDesc:
        ovrEyeType Eye  
        ovrFovPort Fov  
        ovrRecti DistortedViewport  
        ovrVector2f PixelsPerTanAngleAtCenter
        ovrPosef HmdToEyePose  

    ctypedef struct ovrTimewarpProjectionDesc:
        float Projection22 
        float Projection23 
        float Projection32 

    ctypedef struct ovrViewScaleDesc:
        ovrPosef[2] HmdToEyePose
        float HmdSpaceToWorldScaleInMeters

    ctypedef enum ovrTextureType:
        ovrTexture_2D,
        ovrTexture_2D_External, 
        ovrTexture_Cube,
        ovrTexture_Count

    ctypedef enum ovrTextureBindFlags:
        ovrTextureBind_None,
        ovrTextureBind_DX_RenderTarget = 0x0001,
        ovrTextureBind_DX_UnorderedAccess = 0x0002,
        ovrTextureBind_DX_DepthStencil = 0x0004

    ctypedef enum ovrTextureFormat:
        OVR_FORMAT_UNKNOWN = 0,
        OVR_FORMAT_B5G6R5_UNORM = 1, 
        OVR_FORMAT_B5G5R5A1_UNORM = 2, 
        OVR_FORMAT_B4G4R4A4_UNORM = 3, 
        OVR_FORMAT_R8G8B8A8_UNORM = 4,
        OVR_FORMAT_R8G8B8A8_UNORM_SRGB = 5,
        OVR_FORMAT_B8G8R8A8_UNORM = 6,
        OVR_FORMAT_B8G8R8_UNORM = 27,
        OVR_FORMAT_B8G8R8A8_UNORM_SRGB = 7, 
        OVR_FORMAT_B8G8R8X8_UNORM = 8, 
        OVR_FORMAT_B8G8R8X8_UNORM_SRGB = 9, 
        OVR_FORMAT_R16G16B16A16_FLOAT = 10,
        OVR_FORMAT_R11G11B10_FLOAT = 25,
        OVR_FORMAT_D16_UNORM = 11,
        OVR_FORMAT_D24_UNORM_S8_UINT = 12,
        OVR_FORMAT_D32_FLOAT = 13,
        OVR_FORMAT_D32_FLOAT_S8X24_UINT = 14,
        OVR_FORMAT_BC1_UNORM = 15,
        OVR_FORMAT_BC1_UNORM_SRGB = 16,
        OVR_FORMAT_BC2_UNORM = 17,
        OVR_FORMAT_BC2_UNORM_SRGB = 18,
        OVR_FORMAT_BC3_UNORM = 19,
        OVR_FORMAT_BC3_UNORM_SRGB = 20,
        OVR_FORMAT_BC6H_UF16 = 21,
        OVR_FORMAT_BC6H_SF16 = 22,
        OVR_FORMAT_BC7_UNORM = 23,
        OVR_FORMAT_BC7_UNORM_SRGB = 24

    ctypedef enum ovrTextureFlags:
        ovrTextureMisc_None,
        ovrTextureMisc_DX_Typeless = 0x0001,
        ovrTextureMisc_AllowGenerateMips = 0x0002,
        ovrTextureMisc_ProtectedContent = 0x0004,
        ovrTextureMisc_AutoGenerateMips = 0x0008

    ctypedef struct ovrTextureSwapChainDesc:
        ovrTextureType Type
        ovrTextureFormat Format
        int ArraySize 
        int Width
        int Height
        int MipLevels
        int SampleCount
        ovrBool StaticImage 
        unsigned int MiscFlags
        unsigned int BindFlags 

    ctypedef enum ovrMirrorOptions:
        ovrMirrorOption_Default = 0x0000,
        ovrMirrorOption_PostDistortion = 0x0001,
        ovrMirrorOption_LeftEyeOnly = 0x0002,
        ovrMirrorOption_RightEyeOnly = 0x0004,
        ovrMirrorOption_IncludeGuardian = 0x0008,
        ovrMirrorOption_IncludeNotifications = 0x0010,
        ovrMirrorOption_IncludeSystemGui = 0x0020

    ctypedef struct ovrMirrorTextureDesc:
        ovrTextureFormat Format
        int Width
        int Height
        unsigned int MiscFlags 
        unsigned int MirrorOptions

    ctypedef struct ovrTextureSwapChainData
    ctypedef ovrTextureSwapChainData* ovrTextureSwapChain
    ctypedef struct ovrMirrorTextureData
    ctypedef ovrMirrorTextureData* ovrMirrorTexture

    ctypedef enum ovrButton:
        ovrButton_A = 0x00000001,
        ovrButton_B = 0x00000002,
        ovrButton_RThumb = 0x00000004,
        ovrButton_RShoulder = 0x00000008,
        ovrButton_X = 0x00000100,
        ovrButton_Y = 0x00000200,
        ovrButton_LThumb = 0x00000400,
        ovrButton_LShoulder = 0x00000800,
        ovrButton_Up = 0x00010000,
        ovrButton_Down = 0x00020000,
        ovrButton_Left = 0x00040000,
        ovrButton_Right = 0x00080000,
        ovrButton_Enter = 0x00100000,
        ovrButton_Back = 0x00200000,
        ovrButton_VolUp = 0x00400000,
        ovrButton_VolDown = 0x00800000,
        ovrButton_Home = 0x01000000,
        ovrButton_Private = ovrButton_VolUp | ovrButton_VolDown | ovrButton_Home,
        ovrButton_RMask = ovrButton_A | ovrButton_B | ovrButton_RThumb | ovrButton_RShoulder,
        ovrButton_LMask = ovrButton_X | ovrButton_Y | ovrButton_LThumb | ovrButton_LShoulder | ovrButton_Enter

    ctypedef enum ovrTouch:
        ovrTouch_A = ovrButton_A,
        ovrTouch_B = ovrButton_B,
        ovrTouch_RThumb = ovrButton_RThumb,
        ovrTouch_RThumbRest = 0x00000008,
        ovrTouch_RIndexTrigger = 0x00000010,
        ovrTouch_RButtonMask = ovrTouch_A | ovrTouch_B | ovrTouch_RThumb | ovrTouch_RThumbRest | ovrTouch_RIndexTrigger,
        ovrTouch_X = ovrButton_X,
        ovrTouch_Y = ovrButton_Y,
        ovrTouch_LThumb = ovrButton_LThumb,
        ovrTouch_LThumbRest = 0x00000800,
        ovrTouch_LIndexTrigger = 0x00001000,
        ovrTouch_LButtonMask = ovrTouch_X | ovrTouch_Y | ovrTouch_LThumb | ovrTouch_LThumbRest | ovrTouch_LIndexTrigger,
        ovrTouch_RIndexPointing = 0x00000020,
        ovrTouch_RThumbUp = 0x00000040,
        ovrTouch_LIndexPointing = 0x00002000,
        ovrTouch_LThumbUp = 0x00004000,
        ovrTouch_RPoseMask = ovrTouch_RIndexPointing | ovrTouch_RThumbUp,
        ovrTouch_LPoseMask = ovrTouch_LIndexPointing | ovrTouch_LThumbUp

    ctypedef struct ovrTouchHapticsDesc:
        int SampleRateHz
        int SampleSizeInBytes
        int QueueMinSizeToAvoidStarvation
        int SubmitMinSamples
        int SubmitMaxSamples
        int SubmitOptimalSamples

    ctypedef enum ovrControllerType:
        ovrControllerType_None = 0x0000,
        ovrControllerType_LTouch = 0x0001,
        ovrControllerType_RTouch = 0x0002,
        ovrControllerType_Touch = (ovrControllerType_LTouch | ovrControllerType_RTouch),
        ovrControllerType_Remote = 0x0004,
        ovrControllerType_XBox = 0x0010,
        ovrControllerType_Object0 = 0x0100,
        ovrControllerType_Object1 = 0x0200,
        ovrControllerType_Object2 = 0x0400,
        ovrControllerType_Object3 = 0x0800,
        ovrControllerType_Active = 0xffffffff

    ctypedef enum ovrHapticsBufferSubmitMode:
        ovrHapticsBufferSubmit_Enqueue

    cdef int OVR_HAPTICS_BUFFER_SAMPLES_MAX = 256

    ctypedef struct ovrHapticsBuffer:
        const void* Samples
        int SamplesCount
        ovrHapticsBufferSubmitMode SubmitMode

    ctypedef struct ovrHapticsPlaybackState:
        int RemainingQueueSpace
        int SamplesQueued

    ctypedef enum ovrTrackedDeviceType:
        ovrTrackedDevice_None = 0x0000,
        ovrTrackedDevice_HMD = 0x0001,
        ovrTrackedDevice_LTouch = 0x0002,
        ovrTrackedDevice_RTouch = 0x0004,
        ovrTrackedDevice_Touch = (ovrTrackedDevice_LTouch | ovrTrackedDevice_RTouch),
        ovrTrackedDevice_Object0 = 0x0010,
        ovrTrackedDevice_Object1 = 0x0020,
        ovrTrackedDevice_Object2 = 0x0040,
        ovrTrackedDevice_Object3 = 0x0080

    ctypedef enum ovrBoundaryType:
        ovrBoundary_PlayArea = 0x0100

    ctypedef struct ovrBoundaryLookAndFeel:
        ovrColorf Color

    ctypedef struct ovrBoundaryTestResult:
        ovrBool IsTriggering
        float ClosestDistance
        ovrVector3f ClosestPoint
        ovrVector3f ClosestPointNormal

    ctypedef enum ovrHandType:
        ovrHand_Left = 0,
        ovrHand_Right = 1,
        ovrHand_Count = 2

    ctypedef struct ovrInputState:
        double TimeInSeconds
        unsigned int Buttons
        unsigned int Touches
        float[2] IndexTrigger
        float[2] HandTrigger
        ovrVector2f[2] Thumbstick
        ovrControllerType ControllerType
        float[2] IndexTriggerNoDeadzone
        float[2] HandTriggerNoDeadzone
        ovrVector2f[2] ThumbstickNoDeadzone
        float[2] IndexTriggerRaw
        float[2] HandTriggerRaw
        ovrVector2f[2] ThumbstickRaw

    ctypedef struct ovrCameraIntrinsics:
        double LastChangedTime
        ovrFovPort FOVPort
        float VirtualNearPlaneDistanceMeters
        float VirtualFarPlaneDistanceMeters
        ovrSizei ImageSensorPixelResolution
        ovrMatrix4f LensDistortionMatrix
        double ExposurePeriodSeconds
        double ExposureDurationSeconds

    ctypedef enum ovrCameraStatusFlags:
        ovrCameraStatus_None = 0x0,
        ovrCameraStatus_Connected = 0x1,
        ovrCameraStatus_Calibrating = 0x2,
        ovrCameraStatus_CalibrationFailed = 0x4,
        ovrCameraStatus_Calibrated = 0x8,
        ovrCameraStatus_Capturing = 0x10

    ctypedef struct ovrCameraExtrinsics:
        double LastChangedTimeSeconds
        unsigned int CameraStatusFlags
        ovrTrackedDeviceType AttachedToDevice
        ovrPosef RelativePose
        double LastExposureTimeSeconds
        double ExposureLatencySeconds
        double AdditionalLatencySeconds

    ctypedef enum:
        OVR_MAX_EXTERNAL_CAMERA_COUNT = 16
        OVR_EXTERNAL_CAMERA_NAME_SIZE = 32
        
    ctypedef struct ovrExternalCamera:
        char[OVR_EXTERNAL_CAMERA_NAME_SIZE] Name
        ovrCameraIntrinsics Intrinsics
        ovrCameraExtrinsics Extrinsics

    ctypedef enum ovrInitFlags:
        ovrInit_Debug = 0x00000001,
        ovrInit_RequestVersion = 0x00000004,
        ovrInit_Invisible = 0x00000010,
        ovrInit_MixedRendering = 0x00000020,
        ovrInit_FocusAware = 0x00000040,
        ovrinit_WritableBits = 0x00ffffff

    ctypedef enum ovrLogLevel:
        ovrLogLevel_Debug = 0,
        ovrLogLevel_Info = 1, 
        ovrLogLevel_Error = 2 

    ctypedef void(*ovrLogCallback)(uintptr_t userData, int level, const char* message)

    ctypedef struct ovrInitParams:
        uint32_t Flags
        uint32_t RequestedMinorVersion
        ovrLogCallback LogCallback
        uintptr_t UserData
        uint32_t ConnectionTimeoutMS

    cdef ovrResult ovr_Initialize(const ovrInitParams* params)
    cdef void ovr_Shutdown()
    cdef void ovr_GetLastErrorInfo(ovrErrorInfo* errorInfo)
    cdef const char* ovr_GetVersionString()
    cdef int ovr_TraceMessage(int level, const char* message)
    cdef ovrResult ovr_IdentifyClient(const char* identity)
    cdef ovrHmdDesc ovr_GetHmdDesc(ovrSession session)
    cdef unsigned int ovr_GetTrackerCount(ovrSession session)
    cdef ovrTrackerDesc ovr_GetTrackerDesc(ovrSession session, unsigned int trackerDescIndex)
    cdef ovrResult ovr_Create(ovrSession* pSession, ovrGraphicsLuid* pLuid)
    cdef void ovr_Destroy(ovrSession session)

    ctypedef struct ovrSessionStatus:
        ovrBool IsVisible
        ovrBool HmdPresent
        ovrBool HmdMounted
        ovrBool DisplayLost
        ovrBool ShouldQuit
        ovrBool ShouldRecenter
        ovrBool HasInputFocus
        ovrBool OverlayPresent
        ovrBool DepthRequested

    cdef ovrResult ovr_GetSessionStatus(ovrSession session, ovrSessionStatus* sessionStatus)
    cdef ovrResult ovr_IsExtensionSupported(ovrSession session, ovrExtensions extension, ovrBool* outExtensionSupported);
    cdef ovrResult ovr_EnableExtension(ovrSession session, ovrExtensions extension)
    cdef ovrResult ovr_SetTrackingOriginType(ovrSession session, ovrTrackingOrigin origin)
    cdef ovrTrackingOrigin ovr_GetTrackingOriginType(ovrSession session)
    cdef ovrResult ovr_RecenterTrackingOrigin(ovrSession session)
    cdef ovrResult ovr_SpecifyTrackingOrigin(ovrSession session, ovrPosef originPose)
    cdef void ovr_ClearShouldRecenterFlag(ovrSession session)
    cdef ovrTrackingState ovr_GetTrackingState(ovrSession session, double absTime, ovrBool latencyMarker)
    cdef ovrResult ovr_GetDevicePoses(ovrSession session, ovrTrackedDeviceType* deviceTypes, int deviceCount, double absTime, ovrPoseStatef* outDevicePoses)
    cdef ovrTrackerPose ovr_GetTrackerPose(ovrSession session, unsigned int trackerPoseIndex)
    cdef ovrResult ovr_GetInputState(ovrSession session, ovrControllerType controllerType, ovrInputState* inputState)
    cdef unsigned int ovr_GetConnectedControllerTypes(ovrSession session)
    cdef ovrTouchHapticsDesc ovr_GetTouchHapticsDesc(ovrSession session, ovrControllerType controllerType)
    cdef ovrResult ovr_SetControllerVibration(ovrSession session, ovrControllerType controllerType, float frequency, float amplitude)
    cdef ovrResult ovr_SubmitControllerVibration(ovrSession session, ovrControllerType controllerType, const ovrHapticsBuffer* buffer)
    cdef ovrResult ovr_GetControllerVibrationState(ovrSession session, ovrControllerType controllerType, ovrHapticsPlaybackState* outState)
    cdef ovrResult ovr_TestBoundary(ovrSession session, ovrTrackedDeviceType deviceBitmask, ovrBoundaryType boundaryType, ovrBoundaryTestResult* outTestResult)
    cdef ovrResult ovr_TestBoundaryPoint(ovrSession session, const ovrVector3f* point, ovrBoundaryType singleBoundaryType, ovrBoundaryTestResult* outTestResult)
    cdef ovrResult ovr_SetBoundaryLookAndFeel(ovrSession session, const ovrBoundaryLookAndFeel* lookAndFeel)
    cdef ovrResult ovr_ResetBoundaryLookAndFeel(ovrSession session)
    cdef ovrResult ovr_GetBoundaryGeometry(ovrSession session, ovrBoundaryType boundaryType, ovrVector3f* outFloorPoints, int* outFloorPointsCount)
    cdef ovrResult ovr_GetBoundaryDimensions(ovrSession session, ovrBoundaryType boundaryType, ovrVector3f* outDimensions)
    cdef ovrResult ovr_GetBoundaryVisible(ovrSession session, ovrBool* outIsVisible)
    cdef ovrResult ovr_RequestBoundaryVisible(ovrSession session, ovrBool visible)
    cdef ovrResult ovr_GetExternalCameras(ovrSession session, ovrExternalCamera* cameras, unsigned int* inoutCameraCount)
    cdef ovrResult ovr_SetExternalCameraProperties(ovrSession session, const char* name, const ovrCameraIntrinsics* const intrinsics, const ovrCameraExtrinsics* const extrinsics)

    cdef int ovrMaxLayerCount = 16

    ctypedef enum ovrLayerType:
        ovrLayerType_Disabled = 0,
        ovrLayerType_EyeFov = 1,
        ovrLayerType_EyeFovDepth = 2,
        ovrLayerType_Quad = 3,
        ovrLayerType_EyeMatrix = 5,
        ovrLayerType_EyeFovMultires = 7,
        ovrLayerType_Cylinder = 8,
        ovrLayerType_Cube = 10

    ctypedef enum ovrLayerFlags:
        ovrLayerFlag_HighQuality = 0x01,
        ovrLayerFlag_TextureOriginAtBottomLeft = 0x02,
        ovrLayerFlag_HeadLocked = 0x04

    ctypedef struct ovrLayerHeader:
        ovrLayerType Type
        unsigned int Flags

    ctypedef struct ovrLayerEyeFov:
        ovrLayerHeader Header
        ovrTextureSwapChain[2] ColorTexture
        ovrRecti[2] Viewport
        ovrFovPort[2] Fov
        ovrPosef[2] RenderPose
        double SensorSampleTime

    ctypedef struct ovrLayerEyeFovDepth:
        ovrLayerHeader Header
        ovrTextureSwapChain[2] ColorTexture
        ovrRecti[2] Viewport
        ovrFovPort[2] Fov
        ovrPosef[2] RenderPose
        double SensorSampleTime
        ovrTextureSwapChain[2] DepthTexture
        ovrTimewarpProjectionDesc ProjectionDesc

    ctypedef enum ovrTextureLayout:
        ovrTextureLayout_Rectilinear = 0, 
        ovrTextureLayout_Octilinear = 1 

    ctypedef struct ovrTextureLayoutOctilinear:
        float WarpLeft
        float WarpRight
        float WarpUp
        float WarpDown
        float SizeLeft
        float SizeRight
        float SizeUp
        float SizeDown

    ctypedef union ovrTextureLayoutDesc_Union:
        ovrTextureLayoutOctilinear[2] Octilinear

    ctypedef struct ovrLayerEyeFovMultires:
        ovrLayerHeader Header
        ovrTextureSwapChain[2] ColorTexture
        ovrRecti[2] Viewport
        ovrFovPort[2] Fov
        ovrPosef[2] RenderPose
        double SensorSampleTime
        ovrTextureLayout TextureLayout
        ovrTextureLayoutDesc_Union TextureLayoutDesc

    ctypedef struct ovrLayerEyeMatrix:
        ovrLayerHeader Header
        ovrTextureSwapChain[2] ColorTexture
        ovrRecti[2] Viewport
        ovrPosef[2] RenderPose
        ovrMatrix4f[2] Matrix
        double SensorSampleTime

    ctypedef struct ovrLayerQuad:
        ovrLayerHeader Header
        ovrTextureSwapChain ColorTexture
        ovrRecti Viewport
        ovrPosef QuadPoseCenter
        ovrVector2f QuadSize

    ctypedef struct ovrLayerCylinder:
        ovrLayerHeader Header
        ovrTextureSwapChain ColorTexture
        ovrRecti Viewport
        ovrPosef CylinderPoseCenter
        float CylinderRadius
        float CylinderAngle
        float CylinderAspectRatio

    ctypedef struct ovrLayerCube:
        ovrLayerHeader Header
        ovrQuatf Orientation
        ovrTextureSwapChain CubeMapTexture

    ctypedef union ovrLayer_Union:
        ovrLayerHeader Header
        ovrLayerEyeFov EyeFov
        ovrLayerEyeFovDepth EyeFovDepth
        ovrLayerQuad Quad
        ovrLayerEyeFovMultires Multires
        ovrLayerCylinder Cylinder
        ovrLayerCube Cube

    cdef ovrResult ovr_GetTextureSwapChainLength(ovrSession session, ovrTextureSwapChain chain, int* out_Length)
    cdef ovrResult ovr_GetTextureSwapChainCurrentIndex(ovrSession session, ovrTextureSwapChain chain, int* out_Index)
    cdef ovrResult ovr_GetTextureSwapChainDesc(ovrSession session, ovrTextureSwapChain chain, ovrTextureSwapChainDesc* out_Desc)
    cdef ovrResult ovr_CommitTextureSwapChain(ovrSession session, ovrTextureSwapChain chain)
    cdef void ovr_DestroyTextureSwapChain(ovrSession session, ovrTextureSwapChain chain)
    cdef void ovr_DestroyMirrorTexture(ovrSession session, ovrMirrorTexture mirrorTexture)
    cdef ovrSizei ovr_GetFovTextureSize(ovrSession session, ovrEyeType eye, ovrFovPort fov, float pixelsPerDisplayPixel)
    cdef ovrEyeRenderDesc ovr_GetRenderDesc(ovrSession session, ovrEyeType eyeType, ovrFovPort fov)
    cdef ovrResult ovr_WaitToBeginFrame(ovrSession session, long long frameIndex)
    cdef ovrResult ovr_BeginFrame(ovrSession session, long long frameIndex)
    # note edited argument "ovrLayerHeader const* const* layerPtrList"
    cdef ovrResult ovr_EndFrame(ovrSession session, long long frameIndex, const ovrViewScaleDesc* viewScaleDesc, const ovrLayerHeader *const *layerPtrList, unsigned int layerCount)
    cdef ovrResult ovr_SubmitFrame(ovrSession session, long long frameIndex, const ovrViewScaleDesc* viewScaleDesc, const ovrLayerHeader *const *layerPtrList, unsigned int layerCount)

    ctypedef struct ovrPerfStatsPerCompositorFrame:
        int HmdVsyncIndex
        int AppFrameIndex
        int AppDroppedFrameCount
        float AppMotionToPhotonLatency
        float AppQueueAheadTime
        float AppCpuElapsedTime
        float AppGpuElapsedTime
        int CompositorFrameIndex
        int CompositorDroppedFrameCount
        float CompositorLatency
        float CompositorCpuElapsedTime
        float CompositorGpuElapsedTime
        float CompositorCpuStartToGpuEndElapsedTime
        float CompositorGpuEndToVsyncElapsedTime
        ovrBool AswIsActive
        int AswActivatedToggleCount
        int AswPresentedFrameCount
        int AswFailedFrameCount

    cdef int ovrMaxProvidedFrameStats = 5

    ctypedef struct ovrPerfStats:
        ovrPerfStatsPerCompositorFrame[5] FrameStats
        int FrameStatsCount
        ovrBool AnyFrameStatsDropped
        float AdaptiveGpuPerformanceScale
        ovrBool AswIsAvailable
        ovrProcessId VisibleProcessId

    cdef ovrResult ovr_GetPerfStats(ovrSession session, ovrPerfStats* outStats)
    cdef ovrResult ovr_ResetPerfStats(ovrSession session)
    cdef double ovr_GetPredictedDisplayTime(ovrSession session, long long frameIndex)
    cdef double ovr_GetTimeInSeconds()

    ctypedef enum ovrPerfHudMode:
        ovrPerfHud_Off = 0,
        ovrPerfHud_PerfSummary = 1,
        ovrPerfHud_LatencyTiming = 2, 
        ovrPerfHud_AppRenderTiming = 3,
        ovrPerfHud_CompRenderTiming = 4,
        ovrPerfHud_AswStats = 6, 
        ovrPerfHud_VersionInfo = 5, 
        ovrPerfHud_Count = 7 

    ctypedef enum ovrLayerHudMode:
        ovrLayerHud_Off = 0,
        ovrLayerHud_Info = 1

    ctypedef enum ovrDebugHudStereoMode:
        ovrDebugHudStereo_Off = 0,
        ovrDebugHudStereo_Quad = 1,
        ovrDebugHudStereo_QuadWithCrosshair = 2,
        ovrDebugHudStereo_CrosshairAtInfinity = 3,
        ovrDebugHudStereo_Count

    cdef ovrBool ovr_GetBool(ovrSession session, const char* propertyName, ovrBool defaultVal)
    cdef ovrBool ovr_SetBool(ovrSession session, const char* propertyName, ovrBool value)
    cdef int ovr_GetInt(ovrSession session, const char* propertyName, int defaultVal)
    cdef ovrBool ovr_SetInt(ovrSession session, const char* propertyName, int value)
    cdef float ovr_GetFloat(ovrSession session, const char* propertyName, float defaultVal)
    cdef ovrBool ovr_SetFloat(ovrSession session, const char* propertyName, float value)
    cdef unsigned int ovr_GetFloatArray(ovrSession session, const char* propertyName, float values[], unsigned int valuesCapacity)
    cdef ovrBool ovr_SetFloatArray(ovrSession session, const char* propertyName, const float values[], unsigned int valuesSize)
    cdef const char* ovr_GetString(ovrSession session, const char* propertyName, const char* defaultVal)
    cdef ovrBool ovr_SetString(ovrSession session, const char* propertyName, const char* value)


# additional types
ctypedef struct ovrVector4f:
    float x
    float y
    float z
    float w


cimport ovr_math
cimport ovr_capi

cdef class ovrVector2i:
    cdef ovr_capi.ovrVector2i* c_data
    cdef ovr_capi.ovrVector2i  c_ovrVector2i

cdef class ovrSizei:
    cdef ovr_capi.ovrSizei* c_data
    cdef ovr_capi.ovrSizei  c_ovrSizei

cdef class ovrRecti:
    cdef ovr_capi.ovrRecti* c_data
    cdef ovr_capi.ovrRecti  c_ovrRecti

cdef class ovrVector3f:
    cdef ovr_math.Vector3f* c_data
    cdef ovr_math.Vector3f  c_Vector3f

cdef class ovrFovPort:
    cdef ovr_capi.ovrFovPort* c_data
    cdef ovr_capi.ovrFovPort  c_ovrFovPort

cdef class ovrQuatf:
    cdef ovr_math.Quatf* c_data
    cdef ovr_math.Quatf  c_Quatf

cdef class ovrPosef:
    cdef ovr_math.Posef* c_data
    cdef ovr_math.Posef  c_Posef

cdef class ovrMatrix4f:
    cdef ovr_math.Matrix4f* c_data
    cdef ovr_math.Matrix4f  c_Matrix4f


# distutils: language=c++
#  =============================================================================
#  OVR OpenGL Specific (OVR_CAPI_GL.h) Cython Declaration File 
#  =============================================================================
#
#  ovr_capi_gl.pxd
#
#  Copyright 2018 Matthew Cutone <cutonem(a)yorku.ca> and Laurie M. Wilcox
#  <lmwilcox(a)yorku.ca>; The Centre For Vision Research, York University, 
#  Toronto, Canada
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in 
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# 
"""This file exposes Oculus Rift(TM) C API types and functions, allowing Cython 
extensions to access them. The declarations in the file are contemporaneous
with version 1.24 (retrieved 04.15.2018) of the Oculus Rift(TM) PC SDK. 

"""
from ovr_errorcode cimport ovrResult
from ovr_capi cimport ovrSession
from ovr_capi cimport ovrTextureSwapChainDesc, ovrTextureSwapChain
from ovr_capi cimport ovrMirrorTextureDesc, ovrMirrorTexture

cdef extern from "OVR_CAPI_GL.h":
    cdef ovrResult ovr_CreateTextureSwapChainGL(ovrSession session, const ovrTextureSwapChainDesc* desc, ovrTextureSwapChain* out_TextureSwapChain)
    cdef ovrResult ovr_GetTextureSwapChainBufferGL(ovrSession session, ovrTextureSwapChain chain, int index, unsigned int* out_TexId)
    cdef ovrResult ovr_CreateMirrorTextureWithOptionsGL(ovrSession session, const ovrMirrorTextureDesc* desc, ovrMirrorTexture* out_MirrorTexture)
    cdef ovrResult ovr_CreateMirrorTextureGL(ovrSession session, const ovrMirrorTextureDesc* desc, ovrMirrorTexture* out_MirrorTexture)
    cdef ovrResult ovr_GetMirrorTextureBufferGL(ovrSession session, ovrMirrorTexture mirrorTexture, unsigned int* out_TexId)



# distutils: language=c++
from math cimport *
cimport libc.math as cmath
import OpenGL.GL as GL

cdef class ovrVector2i:
    def __cinit__(self, int x=0, int y=0):
        self.c_data = &self.c_ovrVector2i

        self.c_data.x = x
        self.c_data.y = y

    @property
    def x(self):
        return self.c_data.x

    @x.setter
    def x(self, int value):
        self.c_data.x = value

    @property
    def y(self):
        return self.c_data.y

    @y.setter
    def y(self, int value):
        self.c_data.y = value

    def asTuple(self):
        return self.c_data.x, self.c_data.y

    def asList(self):
        return list(self.c_data.x, self.c_data.y)

    def __eq__(self, ovrVector2i b):
        return self.c_data.x == b.c_data.x and self.c_data.y == b.c_data.y

    def __ne__(self, ovrVector2i b):
        return self.c_data.x != b.c_data.x or self.c_data.y != b.c_data.y

    def __add__(ovrVector2i a, ovrVector2i b):
        cdef ovrVector2i to_return = ovrVector2i(
            a.c_data.x + b.c_data.x,
            a.c_data.y + b.c_data.y)

        return to_return

    def __iadd__(self, ovrVector2i b):
        self.c_data.x += b.c_data.x
        self.c_data.y += b.c_data.y

        return self

    def __sub__(ovrVector2i a, ovrVector2i b):
        cdef ovrVector2i to_return = ovrVector2i(
            a.c_data.x - b.c_data.x,
            a.c_data.y - b.c_data.y)

        return to_return

    def __isub__(self, ovrVector2i b):
        self.c_data.x -= b.c_data.x
        self.c_data.y -= b.c_data.y

        return self

    def __neg__(self):
        cdef ovrVector2i to_return = ovrVector2i(-self.c_data.x, -self.c_data.y)

        return to_return

    def __mul__(ovrVector2i a, object b):
        cdef ovrVector2i to_return
        if isinstance(b, ovrVector2i):
            to_return = ovrVector2i(a.c_data.x * b.c_data.x,
                                    a.c_data.y * b.c_data.y)
        elif isinstance(b, (int, float)):
            to_return = ovrVector2i(a.c_data.x * <int>b,
                                    a.c_data.y * <int>b)

        return to_return

    def __imul__(self, object b):
        cdef ovrVector2i to_return
        if isinstance(b, ovrVector2i):
            self.c_data.x *= b.c_data.x
            self.c_data.y *= b.c_data.y
        elif isinstance(b, (int, float)):
            self.c_data.x *= <int>b
            self.c_data.y *= <int>b

        return self

    def __truediv__(ovrVector2i a, object b):
        cdef int rcp = <int>1 / <int>b
        cdef ovrVector2i to_return = ovrVector2i(
            a.c_data.x * rcp,
            a.c_data.y * rcp)

        return to_return

    def __itruediv__(self, object b):
        cdef int rcp = <int>1 / <int>b
        self.c_data.x *= rcp
        self.c_data.y *= rcp

        return self

    @staticmethod
    def min(ovrVector2i a, ovrVector2i b):
        cdef ovrVector2i to_return = ovrVector2i(
            a.c_data.x if a.c_data.x < b.c_data.x else b.c_data.x,
            a.c_data.y if a.c_data.y < b.c_data.y else b.c_data.y)

        return to_return

    @staticmethod
    def max(ovrVector2i a, ovrVector2i b):
        cdef ovrVector2i to_return = ovrVector2i(
            a.c_data.x if a.c_data.x > b.c_data.x else b.c_data.x,
            a.c_data.y if a.c_data.y > b.c_data.y else b.c_data.y)

        return to_return

    def clamped(self, int max_mag):
        cdef int mag_squared = self.length_sq()
        if mag_squared > max_mag * max_mag:
            return self * (max_mag / cmath.sqrt(mag_squared))

        return self

    def isEqual(self, ovrVector2i b, int tolerance = 0):
        return cmath.fabs(b.c_data.x - self.c_data.x) <= tolerance and \
            cmath.fabs(b.c_data.y - self.c_data.y) <= tolerance

    def compare(self, ovrVector2i b, int tolerance = 0):
        return self.is_equal(b, tolerance)

    def __getitem__(self, int idx):
        assert 0 <= idx < 2
        cdef int* ptr_val = &self.c_data.x + idx

        return <int>ptr_val[0]

    def __setitem__(self, int idx, int val):
        assert 0 <= idx < 2
        cdef int* ptr_val = &self.c_data.x + idx
        ptr_val[0] = val

    def entrywiseMultiply(self, ovrVector2i b):
        cdef ovrVector2i to_return = ovrVector2i(
            self.c_data.x * b.c_data.x,
            self.c_data.y * b.c_data.y)

        return to_return

    def dot(self, ovrVector2i b):
        cdef int dot_prod = \
            self.c_data.x * b.c_data.x + self.c_data.y * b.c_data.y

        return <int>dot_prod

    def angle(self, ovrVector2i b):
        cdef int div = self.length_sq() * b.length_sq()
        assert div != <int>0
        cdef int to_return = self.dot(b) / cmath.sqrt(div)

        return to_return

    def lengthSq(self):
        return \
            <int>(self.c_data.x * self.c_data.x + self.c_data.y * self.c_data.y)

    def length(self):
        return <int>cmath.sqrt(self.length_sq())

    def distanceSq(self, ovrVector2i b):
        return (self - b).length_sq()

    def distance(self, ovrVector2i b):
        return (self - b).length()

    def isNormalized(self):
        return cmath.fabs(self.length_sq() - <int>1) < 0

    def normalize(self):
        cdef int s = self.length()
        if s != <int>0:
            s = <int>1 / s

        self *= s

    def normalized(self):
        cdef int s = self.length()
        if s != <int>0:
            s = <int>1 / s

        return self * s

    def lerp(self, ovrVector2i b, int f):
        return self * (<int>1 - f) + b * f

    def projectTo(self, ovrVector2i b):
        cdef int l2 = self.length_sq()
        assert l2 != <int>0

        return b * (self.dot(b) / l2)

    def isClockwise(self, ovrVector2i b):
        return (self.c_data.x * b.c_data.y - self.c_data.y * b.c_data.x) < 0


cdef class ovrSizei:
    def __cinit__(self, int w=0, int h=0):
        self.c_data = &self.c_ovrSizei

        self.c_data.w = w
        self.c_data.h = h

    @property
    def w(self):
        return self.c_data.w

    @w.setter
    def w(self, int value):
        self.c_data.w = value

    @property
    def h(self):
        return self.c_data.h

    @h.setter
    def h(self, int value):
        self.c_data.h = value

    def asTuple(self):
        return self.c_data.w, self.c_data.h

    def asList(self):
        return [self.c_data.w, self.c_data.h]

    def __eq__(self, ovrSizei b):
        return self.c_data.w == b.c_data.w and self.c_data.h == b.c_data.h

    def __ne__(self, ovrSizei b):
        return self.c_data.w != b.c_data.w or self.c_data.h != b.c_data.h

    def __add__(ovrSizei a, ovrSizei b):
        cdef ovrSizei to_return = ovrSizei(
            a.c_data.w + b.c_data.w,
            a.c_data.h + b.c_data.h)

        return to_return

    def __iadd__(self, ovrSizei b):
        self.c_data.w += b.c_data.w
        self.c_data.h += b.c_data.h

        return self

    def __sub__(ovrSizei a, ovrSizei b):
        cdef ovrSizei to_return = ovrSizei(
            a.c_data.w - b.c_data.w,
            a.c_data.h - b.c_data.h)

        return to_return

    def __isub__(self, ovrSizei b):
        self.c_data.w -= b.c_data.w
        self.c_data.h -= b.c_data.h

        return self

    def __neg__(self):
        cdef ovrSizei to_return = ovrSizei(-self.c_data.w, -self.c_data.h)

        return to_return

    def __mul__(ovrSizei a, object b):
        cdef ovrSizei to_return
        if isinstance(b, ovrSizei):
            to_return = ovrSizei(a.c_data.w * b.c_data.w,
                                 a.c_data.h * b.c_data.h)
        elif isinstance(b, (int, float)):
            to_return = ovrSizei(a.c_data.w * <int>b,
                                 a.c_data.h * <int>b)

        return to_return

    def __imul__(self, object b):
        cdef ovrSizei to_return
        if isinstance(b, ovrSizei):
            self.c_data.w *= b.c_data.w
            self.c_data.h *= b.c_data.h
        elif isinstance(b, (int, float)):
            self.c_data.w *= <int>b
            self.c_data.h *= <int>b

        return self

    def __truediv__(ovrSizei a, object b):
        cdef float rcp = <float>1 / <float>b
        cdef ovrSizei to_return = ovrSizei(
            <int>(<float>a.c_data.w * rcp),
            <int>(<float>a.c_data.h * rcp))

        return to_return

    def __itruediv__(self, object b):
        cdef float rcp = <float>1 / <float>b
        self.c_data.w = <int>(<float>self.c_data.w * rcp)
        self.c_data.h = <int>(<float>self.c_data.h * rcp)

        return self

    @staticmethod
    def min(ovrSizei a, ovrSizei b):
        cdef ovrSizei to_return = ovrSizei(
            a.c_data.w if a.c_data.w < b.c_data.w else b.c_data.w,
            a.c_data.h if a.c_data.h < b.c_data.h else b.c_data.h)

        return to_return

    @staticmethod
    def max(ovrSizei a, ovrSizei b):
        cdef ovrSizei to_return = ovrSizei(
            a.c_data.w if a.c_data.w > b.c_data.w else b.c_data.w,
            a.c_data.h if a.c_data.h > b.c_data.h else b.c_data.h)

        return to_return

    def area(self):
        return self.c_data.w * self.c_data.h

    def toVector(self):
        cdef ovrVector2i to_return = ovrVector2i(self.c_data.w, self.c_data.h)

        return to_return

    def asVector(self):
        return self.to_vector()


cdef class ovrRecti:

    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, *args, **kwargs):
        self.c_data = &self.c_ovrRecti

        cdef int nargin = <int>len(args)  # get number of arguments
        if nargin == 0:
            self.c_data.Pos.x = 0
            self.c_data.Pos.y = 0
            self.c_data.Size.w = 0
            self.c_data.Size.h = 0
        elif nargin == 4:
            self.c_data.Pos.x = args[0]
            self.c_data.Pos.y = args[1]
            self.c_data.Size.w = args[2]
            self.c_data.Size.h = args[3]
        elif nargin == 2 and \
                isinstance(args[0], ovrVector2i) and \
                isinstance(args[1], ovrSizei):
            self.c_data.Pos.x = (<ovrVector2i>args[0]).c_data.x
            self.c_data.Pos.y = (<ovrVector2i>args[0]).c_data.y
            self.c_data.Size.w = (<ovrSizei>args[1]).c_data.w
            self.c_data.Size.h = (<ovrSizei>args[1]).c_data.h

    @property
    def x(self):
        return self.c_data.Pos.x

    @x.setter
    def x(self, int value):
        self.c_data.Pos.x = value

    @property
    def y(self):
        return self.c_data.Pos.y

    @y.setter
    def y(self, int value):
        self.c_data.Pos.y = value

    @property
    def w(self):
        return self.c_data.Size.w

    @w.setter
    def w(self, int value):
        self.c_data.Size.w = value

    @property
    def h(self):
        return self.c_data.Size.h

    @h.setter
    def h(self, int value):
        self.c_data.Size.h = value

    def asTuple(self):
        return self.c_data.Pos.x, self.c_data.Pos.y, \
               self.c_data.Size.w, self.c_data.Size.h

    def asList(self):
        return [self.c_data.Pos.x, self.c_data.Pos.y,
                self.c_data.Size.w, self.c_data.Size.h]

    def __len__(self):
        return 4

    def getPos(self):
        cdef ovrVector2i to_return = ovrVector2i(self.c_data.Pos.x,
                                                 self.c_data.Pos.y)

        return to_return

    def getSize(self):
        cdef ovrSizei to_return = ovrSizei(self.c_data.Size.w,
                                           self.c_data.Size.h)

        return to_return

    def setPos(self, ovrVector2i pos):
        self.c_data.Pos.x = pos.c_data.x
        self.c_data.Pos.y = pos.c_data.y

    def setSize(self, ovrSizei size):
        self.c_data.Size.w = size.c_data.w
        self.c_data.Size.h = size.c_data.h

    def __eq__(self, ovrRecti b):
        return self.c_data.Pos.x == b.c_data.Pos.x and \
               self.c_data.Pos.y == b.c_data.Pos.y and \
               self.c_data.Size.w == b.c_data.Size.w and \
               self.c_data.Size.h == b.c_data.Size.h

    def __ne__(self, ovrRecti b):
        return not self.__eq__(b)


cdef class ovrVector3f:
    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, *args, **kwargs):
        self.c_data = &self.c_Vector3f

        cdef int nargin = <int>len(args)  # get number of arguments
        if nargin == 0:
            self.c_data[0] = ovr_math.Vector3f.Zero()
        elif nargin == 3:
            self.c_data[0] = ovr_math.Vector3f(
                <float>args[0], <float>args[1], <float>args[2])

    @property
    def x(self):
        return self.c_data.x

    @x.setter
    def x(self, float value):
        self.c_data.x = value

    @property
    def y(self):
        return self.c_data.y

    @y.setter
    def y(self, float value):
        self.c_data.y = value

    @property
    def z(self):
        return self.c_data.z

    @z.setter
    def z(self, float value):
        self.c_data.z = value

    def as_tuple(self):
        return self.c_data.x, self.c_data.y, self.c_data.z

    @property
    def ctypes(self):
        return (GL.GLfloat * 3)(
            self.c_data.x,
            self.c_data.y,
            self.c_data.z)

    def __len__(self):
        return 3

    @staticmethod
    def zero():
        cdef ovrVector3f to_return = ovrVector3f()
        return to_return

    def __eq__(self, ovrVector3f b):
        return (<ovrVector3f>self).c_data[0] == b.c_data[0]

    def __ne__(self, ovrVector3f b):
        return (<ovrVector3f>self).c_data[0] != b.c_data[0]

    def __add__(ovrVector3f a, ovrVector3f b):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = a.c_data[0] + b.c_data[0]

        return to_return

    def __iadd__(self, ovrVector3f b):
        (<ovrVector3f>self).c_data[0] += b.c_data[0]

        return self

    def __sub__(ovrVector3f a, ovrVector3f b):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = a.c_data[0] - b.c_data[0]

        return to_return

    def __isub__(self, ovrVector3f b):
        (<ovrVector3f>self).c_data[0] -= b.c_data[0]

        return self

    def __neg__(self):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = -(<ovrVector3f>self).c_data[0]

        return to_return

    def __mul__(ovrVector3f a, object b):
        cdef ovrVector3f to_return = ovrVector3f()
        if isinstance(b, ovrVector3f):
            (<ovrVector3f>to_return).c_data[0] = \
                (<ovrVector3f>a).c_data[0] * (<ovrVector3f>b).c_data[0]
        elif isinstance(b, (int, float)):
            (<ovrVector3f>to_return).c_data[0] = \
                (<ovrVector3f>a).c_data[0] * <float>b

        return to_return

    def __imul__(self, object b):
        if isinstance(b, (int, float)):
            (<ovrVector3f>self).c_data[0] = \
                (<ovrVector3f>self).c_data[0] * <float>b

        return self

    def __truediv__(ovrVector3f a, object b):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrVector3f>a).c_data[0] / (<ovrVector3f>b).c_data[0]

        return to_return

    def __itruediv__(self, object b):
        if isinstance(b, (int, float)):
            (<ovrVector3f>self).c_data[0] = \
                (<ovrVector3f>self).c_data[0] / <float>b

        return self

    @staticmethod
    def min(ovrVector3f a, ovrVector3f b):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            ovr_math.Vector3f.Min(a.c_data[0], b.c_data[0])

        return to_return

    @staticmethod
    def max(ovrVector3f a, ovrVector3f b):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            ovr_math.Vector3f.Max(a.c_data[0], b.c_data[0])

        return to_return

    def clamped(self, float max_mag):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrVector3f>self).c_data[0].Clamped(max_mag)

        return to_return

    def isEqual(self, ovrVector3f b, float tolerance = 0.0):
        return (<ovrVector3f>self).c_data[0].IsEqual(b.c_data[0], tolerance)

    def compare(self, ovrVector3f b, float tolerance = 0.0):
        return self.is_equal(b, tolerance)

    def __getitem__(self, int idx):
        assert 0 <= idx < 3
        cdef float* ptr_val = &self.c_data.x + idx

        return <float>ptr_val[0]

    def __setitem__(self, int idx, float val):
        assert 0 <= idx < 3
        cdef float* ptr_val = &self.c_data.x + idx
        ptr_val[0] = val

    def entrywiseMultiply(self, ovrVector3f b):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrVector3f>self).c_data[0].EntrywiseMultiply(
            b.c_data[0])

        return to_return

    def dot(self, ovrVector3f b):
        return <float>(<ovrVector3f>self).c_data[0].Dot(b.c_data[0])

    def cross(self, ovrVector3f b):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrVector3f>self).c_data[0].Cross(b.c_data[0])

        return to_return

    def angle(self, ovrVector3f b):
        return <float>(<ovrVector3f>self).c_data[0].Angle(b.c_data[0])

    def lengthSq(self):
        return <float>(<ovrVector3f>self).c_data[0].LengthSq()

    def length(self):
        return <float>(<ovrVector3f>self).c_data[0].Length()

    def distanceSq(self, ovrVector3f b):
        return <float>(<ovrVector3f>self).c_data[0].DistanceSq(b.c_data[0])

    def distance(self, ovrVector3f b):
        return <float>(<ovrVector3f>self).c_data[0].Distance(b.c_data[0])

    def isNormalized(self):
        return <bint>(<ovrVector3f>self).c_data[0].IsNormalized()

    def normalize(self):
        (<ovrVector3f>self).c_data[0].Normalize()

    def normalized(self):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrVector3f>self).c_data[0].Normalized()

        return to_return

    def lerp(self, ovrVector3f b, float f):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrVector3f>self).c_data[0].Lerp(b.c_data[0], f)

        return to_return

    def projectTo(self, ovrVector3f b):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrVector3f>self).c_data[0].ProjectTo(
            b.c_data[0])

        return to_return

    def projectToPlane(self, ovrVector3f normal):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrVector3f>self).c_data[0].ProjectToPlane(
                normal.c_data[0])

        return to_return


cdef class ovrFovPort:
    def __cinit__(self,
                  float up_tan=0.0,
                  float down_tan=0.0,
                  float left_tan=0.0,
                  float right_tan=0.0):
        self.c_data = &self.c_ovrFovPort

    @property
    def UpTan(self):
        return <float>self.c_data[0].UpTan

    @UpTan.setter
    def UpTan(self, float value):
        self.c_data[0].UpTan = value

    @property
    def DownTan(self):
        return <float>self.c_data[0].DownTan

    @DownTan.setter
    def DownTan(self, float value):
        self.c_data[0].DownTan = value

    @property
    def LeftTan(self):
        return <float>self.c_data[0].LeftTan

    @LeftTan.setter
    def LeftTan(self, float value):
        self.c_data[0].LeftTan = value

    @property
    def RightTan(self):
        return <float>self.c_data[0].RightTan

    @RightTan.setter
    def RightTan(self, float value):
        self.c_data[0].RightTan = value

    @staticmethod
    def max(ovrFovPort a, ovrFovPort b):
        cdef ovrFovPort to_return = ovrFovPort()
        (<ovrFovPort>to_return).c_data[0].UpTan = cmath.fmax(
            a.c_data[0].UpTan, b.c_data[0].UpTan)
        (<ovrFovPort>to_return).c_data[0].DownTan = cmath.fmax(
            a.c_data[0].DownTan, b.c_data[0].DownTan)
        (<ovrFovPort>to_return).c_data[0].LeftTan = cmath.fmax(
            a.c_data[0].LeftTan, b.c_data[0].LeftTan)
        (<ovrFovPort>to_return).c_data[0].RightTan = cmath.fmax(
            a.c_data[0].RightTan, b.c_data[0].RightTan)

        return to_return

    @staticmethod
    def min(ovrFovPort a, ovrFovPort b):
        cdef ovrFovPort to_return = ovrFovPort()
        (<ovrFovPort>to_return).c_data[0].UpTan = cmath.fmin(
            a.c_data[0].UpTan, b.c_data[0].UpTan)
        (<ovrFovPort>to_return).c_data[0].DownTan = cmath.fmin(
            a.c_data[0].DownTan, b.c_data[0].DownTan)
        (<ovrFovPort>to_return).c_data[0].LeftTan = cmath.fmin(
            a.c_data[0].LeftTan, b.c_data[0].LeftTan)
        (<ovrFovPort>to_return).c_data[0].RightTan = cmath.fmin(
            a.c_data[0].RightTan, b.c_data[0].RightTan)

        return to_return

cdef class ovrQuatf:
    def __cinit__(self, *args):
        self.c_data = &self.c_Quatf

        cdef int nargin = <int>len(args)  # get number of arguments
        if nargin == 0:
            self.c_data[0] = ovr_math.Quatf.Identity()
        elif nargin == 1:
            if isinstance(ovrMatrix4f, args[0]):
                self.c_data[0] = ovr_math.Quatf(
                    (<ovrMatrix4f>args[0]).c_data[0])
            elif isinstance(ovrMatrix4f, args[0]):
                self.c_data[0] = ovr_math.Quatf(
                    (<ovrQuatf>args[0]).c_data[0])
        elif nargin == 2:
            # quaternion from axis and angle
            if isinstance(args[0], ovrVector3f) and \
                    isinstance(args[1], (int, float)):
                self.c_data[0] = ovr_math.Quatf(
                    (<ovrVector3f>args[0]).c_data[0], <float>args[1])
        elif nargin == 4:
            self.c_data[0] = ovr_math.Quatf(
                <float>args[0],
                <float>args[1],
                <float>args[2],
                <float>args[3])

    @property
    def x(self):
        return self.c_data.x

    @x.setter
    def x(self, float value):
        self.c_data.x = value

    @property
    def y(self):
        return self.c_data.y

    @y.setter
    def y(self, float value):
        self.c_data.y = value

    @property
    def z(self):
        return self.c_data.z

    @z.setter
    def z(self, float value):
        self.c_data.z = value

    @property
    def w(self):
        return self.c_data.w

    @w.setter
    def w(self, float value):
        self.c_data.w = value

    def asTuple(self):
        return self.c_data.x, self.c_data.y, self.c_data.z, self.c_data.w

    def asList(self):
        return [self.c_data.x, self.c_data.y, self.c_data.z, self.c_data.w]

    def __len__(self):
        return 4

    def __neg__(self):
        cdef ovrQuatf to_return = ovrQuatf(
            -self.c_data.x, -self.c_data.y, -self.c_data.z, -self.c_data.w)

        return to_return

    @staticmethod
    def identity():
        cdef ovrQuatf to_return = ovrQuatf(0.0, 0.0, 0.0, 1.0)
        return to_return

    def __eq__(self, ovrQuatf b):
        return (<ovrQuatf>self).c_data[0] == b.c_data[0]

    def __ne__(self, ovrQuatf b):
        return (<ovrQuatf>self).c_data[0] != b.c_data[0]

    def __add__(ovrQuatf a, ovrQuatf b):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = a.c_data[0] + b.c_data[0]

        return to_return

    def __iadd__(self, ovrQuatf b):
        self.c_data.x += b.c_data.x
        self.c_data.y += b.c_data.y
        self.c_data.z += b.c_data.z
        self.c_data.w += b.c_data.w

        return self

    def __sub__(ovrQuatf a, ovrQuatf b):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = a.c_data[0] - b.c_data[0]

        return to_return

    def __isub__(self, ovrQuatf b):
        self.c_data.x -= b.c_data.x
        self.c_data.y -= b.c_data.y
        self.c_data.z -= b.c_data.z
        self.c_data.w -= b.c_data.w

        return self

    def __mul__(ovrQuatf a, object b):
        if isinstance(b, ovrVector3f):
            return a.rotate(<ovrVector3f>b)
        elif isinstance(b, ovrQuatf):
            # quaternion multiplication
            return ovrQuatf(
                a.c_data.w * (<ovrQuatf>b).c_data.x +
                a.c_data.x * (<ovrQuatf>b).c_data.w +
                a.c_data.y * (<ovrQuatf>b).c_data.z -
                a.c_data.z * (<ovrQuatf>b).c_data.y,
                a.c_data.w * (<ovrQuatf>b).c_data.y -
                a.c_data.x * (<ovrQuatf>b).c_data.z +
                a.c_data.y * (<ovrQuatf>b).c_data.w +
                a.c_data.z * (<ovrQuatf>b).c_data.x,
                a.c_data.w * (<ovrQuatf>b).c_data.z +
                a.c_data.x * (<ovrQuatf>b).c_data.y -
                a.c_data.y * (<ovrQuatf>b).c_data.x +
                a.c_data.z * (<ovrQuatf>b).c_data.w,
                a.c_data.w * (<ovrQuatf>b).c_data.w -
                a.c_data.x * (<ovrQuatf>b).c_data.x -
                a.c_data.y * (<ovrQuatf>b).c_data.y -
                a.c_data.z * (<ovrQuatf>b).c_data.z)
        elif isinstance(b, (int, float)):
            return ovrQuatf(
                a.c_data.x * <float>b,
                a.c_data.y * <float>b,
                a.c_data.z * <float>b,
                a.c_data.w * <float>b)

    def __imul__(self, object b):
        if isinstance(b, ovrQuatf):
            self.c_data.x *= b.c_data.x
            self.c_data.y *= b.c_data.y
            self.c_data.z *= b.c_data.z
            self.c_data.w *= b.c_data.w
        elif isinstance(b, (int, float)):
            self.c_data.x *= <float>b
            self.c_data.y *= <float>b
            self.c_data.z *= <float>b
            self.c_data.w *= <float>b

        return self

    def __truediv__(ovrQuatf a, float s):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = a.c_data[0] / s

        return to_return

    def __itruediv__(self, float s):
        (<ovrQuatf>self).c_data[0] = (<ovrQuatf>self).c_data[0] / s

        return self

    def isEqual(self, ovrQuatf b, float tolerance = 0.0):
        return  self.abs(self.dot(b)) >= <float>1 - tolerance

    def abs(self, float v):
        return <float>(<ovrQuatf>self).c_data[0].Abs(v)

    def imag(self):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrQuatf>self).c_data[0].Imag()

        return to_return

    def lengthSq(self):
        return <float>(<ovrQuatf>self).c_data[0].LengthSq()

    def length(self):
        return <float>(<ovrQuatf>self).c_data[0].Length()

    def distance(self, ovrQuatf q):
        return <float>(<ovrQuatf>self).c_data[0].Distance(q.c_data[0])

    def distanceSq(self, ovrQuatf q):
        return <float>(<ovrQuatf>self).c_data[0].DistanceSq(q.c_data[0])

    def dot(self, ovrQuatf q):
        return <float>(<ovrQuatf>self).c_data[0].Dot(q.c_data[0])

    def angle(self, ovrQuatf q):
        return <float>(<ovrQuatf>self).c_data[0].Angle(q.c_data[0])

    def isNormalized(self):
        return <bint>(<ovrQuatf>self).c_data[0].IsNormalized()

    def normalize(self):
        (<ovrQuatf>self).c_data[0].Normalize()

        return self

    def normalized(self):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = \
            (<ovrQuatf>self).c_data[0].Normalized()

        return to_return

    def conj(self):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = (<ovrQuatf>self).c_data[0].Conj()

        return to_return

    @staticmethod
    def align(ovrVector3f align_to, ovrVector3f v):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = ovr_math.Quatf.Align(
            align_to.c_data[0], v.c_data[0])

        return to_return

    def rotate(self, ovrVector3f v):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrQuatf>self).c_data[0].Rotate(v.c_data[0])

        return to_return

    def inverseRotate(self, ovrVector3f v):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovrQuatf>self).c_data[0].InverseRotate(v.c_data[0])

        return to_return

    def inverted(self):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = (<ovrQuatf>self).c_data[0].Inverted()

        return to_return

    def inverse(self):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = (<ovrQuatf>self).c_data[0].Inverse()

        return to_return

    def invert(self):
        (<ovrQuatf>self).c_data[0].Inverse()

        return self

    def __invert__(self):
        return self.inverse()


cdef class ovrPosef:
    """ovrPosef

    Class to represent a pose using a vector and quaternion.

    """
    def __cinit__(self, *args, **kwargs):
        self.c_data = &self.c_Posef

        cdef int nargin = <int>len(args)  # get number of arguments
        if nargin == 0:
            self.c_data[0].Translation = ovr_math.Vector3f(0.0, 0.0, 0.0)
            self.c_data[0].Rotation = ovr_math.Quatf.Identity()
        elif nargin == 1:
            if isinstance(args[0], ovrVector3f):
                self.c_data[0].Translation = (<ovrVector3f>args[0]).c_data[0]
            elif isinstance(args[0], ovrQuatf):
                self.c_data[0].Rotation = (<ovrQuatf>args[0]).c_data[0]
        elif nargin == 2:
            if isinstance(args[0], ovrVector3f) and \
                    isinstance(args[1], ovrQuatf):
                self.c_data[0].Translation = (<ovrVector3f>args[0]).c_data[0]
                self.c_data[0].Rotation = (<ovrQuatf>args[1]).c_data[0]

    @property
    def rotation(self):
        cdef ovrQuatf to_return = ovrQuatf()
        (<ovrQuatf>to_return).c_data[0] = self.c_data[0].Rotation

        return to_return

    @rotation.setter
    def rotation(self, ovrQuatf value):
        (<ovrPosef>self).c_data[0].Rotation = value.c_data[0]

    @property
    def translation(self):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = self.c_data[0].Translation

        return to_return

    @translation.setter
    def translation(self, ovrVector3f value):
        (<ovrPosef>self).c_data[0].Translation = value.c_data[0]

    def rotate(self, ovrVector3f v):
        return self.rotation.rotate(v)

    def inverseRotate(self, ovrVector3f v):
        return self.rotation.inverseRotate(v)

    def translate(self, ovrVector3f v):
        return v + self.translation

    def transform(self, ovrVector3f v):
        return self.rotate(v) + self.translation

    def inverseTransform(self, ovrVector3f v):
        return self.inverseRotate(v - self.translation)

    def transformNormal(self, ovrVector3f v):
        return self.rotate(v)

    def inverseTransformNormal(self, ovrVector3f v):
        return self.inverseRotate(v)

    def apply(self, ovrVector3f v):
        return self.transform(v)

    def __mul__(ovrPosef a, ovrPosef b):
        cdef ovrPosef to_return = ovrPosef(
            a.apply(b.translation),
            <ovrQuatf>a.rotation * <ovrQuatf>b.rotation)

        return to_return

    def inverted(self):
        cdef ovrQuatf inv = self.rotation.inverted()
        cdef ovrPosef to_return = ovrPosef(inv, inv.rotate(-self.translation))

    def normalized(self):
        cdef ovrPosef to_return = ovrPosef(
            self.rotation.normalized(),
            self.translation)

        return to_return

    def normalize(self):
        self.rotation.normalize()


cdef class ovrMatrix4f:
    """ovrMatrix4f

    4x4 Matrix typically used for 3D transformations. By default, all matrices
    are right handed. Values are stored in row-major order. Transformations
    are applied left-to-right.

    """
    def __cinit__(self, *args):
        self.c_data = &self.c_Matrix4f

        cdef int nargin = <int>len(args)
        if nargin == 0:
            self.c_data[0] = ovr_math.Matrix4f.Identity()
        elif nargin == 1:
            if isinstance(args[0], ovrQuatf):
                self.c_data[0] = ovr_math.Matrix4f(
                    <ovr_math.Quatf>(<ovrQuatf>args[0]).c_data[0])
            elif isinstance(args[0], ovrPosef):
                self.c_data[0] = ovr_math.Matrix4f(
                    <ovr_math.Posef>(<ovrPosef>args[0]).c_data[0])
            elif isinstance(args[0], ovrMatrix4f):
                self.c_data[0] = ovr_math.Matrix4f(
                    <ovr_math.Matrix4f>(<ovrMatrix4f>args[0]).c_data[0])
        elif nargin == 9:
            self.c_data[0] = ovr_math.Matrix4f(
                <float>args[0],
                <float>args[1],
                <float>args[2],
                <float>args[3],
                <float>args[4],
                <float>args[5],
                <float>args[6],
                <float>args[7],
                <float>args[8])
        elif nargin == 16:
            self.c_data[0] = ovr_math.Matrix4f(
                <float>args[0],
                <float>args[1],
                <float>args[2],
                <float>args[3],
                <float>args[4],
                <float>args[5],
                <float>args[6],
                <float>args[7],
                <float>args[8],
                <float>args[9],
                <float>args[10],
                <float>args[11],
                <float>args[12],
                <float>args[13],
                <float>args[14],
                <float>args[15])

    @property
    def M(self):
        return self.c_data.M

    @staticmethod
    def identity():
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        return to_return

    def setIdentity(self):
        self.c_data[0] = ovr_math.Matrix4f()

    def setXBasis(self, ovrVector3f v):
        (<ovrMatrix4f>self).c_data[0].SetXBasis(v.c_data[0])

    def getXBasis(self):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).GetXBasis()

        return to_return

    @property
    def xBasis(self):
        return self.getXBasis()

    @xBasis.setter
    def xBasis(self, object v):
        if isinstance(v, ovrVector3f):
            self.setXBasis(<ovrVector3f>v)
        elif isinstance(v, (list, tuple)):
            self.setXBasis(ovrVector3f(<float>v[0], <float>v[1], <float>v[2]))

    def setYBasis(self, ovrVector3f v):
        (<ovrMatrix4f>self).c_data[0].SetYBasis(v.c_data[0])

    def getYBasis(self):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).GetYBasis()

        return to_return

    @property
    def yBasis(self):
        return self.getYBasis()

    @yBasis.setter
    def yBasis(self, object v):
        if isinstance(v, ovrVector3f):
            self.setYBasis(<ovrVector3f>v)
        elif isinstance(v, (list, tuple)):
            self.setYBasis(ovrVector3f(<float>v[0], <float>v[1], <float>v[2]))

    def setZBasis(self, ovrVector3f v):
        (<ovrMatrix4f>self).c_data[0].SetZBasis(v.c_data[0])

    def getZBasis(self):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).GetZBasis()

        return to_return

    @property
    def zBasis(self):
        return self.getZBasis()

    @zBasis.setter
    def zBasis(self, object v):
        if isinstance(v, ovrVector3f):
            self.setZBasis(<ovrVector3f>v)
        elif isinstance(v, (list, tuple)):
            self.setZBasis(ovrVector3f(<float>v[0], <float>v[1], <float>v[2]))

    @property
    def ctypes(self):
        return (GL.GLfloat * 16)(
            self.c_data.M[0][0],
            self.c_data.M[1][0],
            self.c_data.M[2][0],
            self.c_data.M[3][0],
            self.c_data.M[0][1],
            self.c_data.M[1][1],
            self.c_data.M[2][1],
            self.c_data.M[3][1],
            self.c_data.M[0][2],
            self.c_data.M[1][2],
            self.c_data.M[2][2],
            self.c_data.M[3][2],
            self.c_data.M[0][3],
            self.c_data.M[1][3],
            self.c_data.M[2][3],
            self.c_data.M[3][3])

    def __eq__(self, ovrMatrix4f b):
        return (<ovrMatrix4f>self).c_data[0] == b.c_data[0]

    def __ne__(self, ovrMatrix4f b):
        return not (<ovrMatrix4f>self).c_data[0] == b.c_data[0]

    def __add__(ovrMatrix4f a, ovrMatrix4f b):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = a.c_data[0] + b.c_data[0]

        return to_return

    def __iadd__(self, ovrMatrix4f b):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>self).c_data[0] = \
            (<ovrMatrix4f>self).c_data[0] + b.c_data[0]

        return self

    def __sub__(ovrMatrix4f a, ovrMatrix4f b):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = a.c_data[0] - b.c_data[0]

        return to_return

    def __isub__(self, ovrMatrix4f b):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>self).c_data[0] = \
            (<ovrMatrix4f>self).c_data[0] - b.c_data[0]

        return self

    @staticmethod
    def multiply(ovrMatrix4f d, ovrMatrix4f a, ovrMatrix4f b):
        ovr_math.Matrix4f.Multiply(&d.c_data[0], a.c_data[0], b.c_data[0])

    def __mul__(ovrMatrix4f a, object b):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        if isinstance(b, ovrMatrix4f):
            (<ovrMatrix4f>to_return).c_data[0] = \
                (<ovrMatrix4f>a).c_data[0] * (<ovrMatrix4f>b).c_data[0]
        elif isinstance(b, (int, float)):
            (<ovrMatrix4f>to_return).c_data[0] = \
                (<ovrMatrix4f>a).c_data[0] * <float>b

        return to_return

    def __imul__(self, object b):
        if isinstance(b, ovrMatrix4f):
            (<ovrMatrix4f>self).c_data[0] = \
                (<ovrMatrix4f>self).c_data[0] * (<ovrMatrix4f>b).c_data[0]
        elif isinstance(b, (int, float)):
            (<ovrMatrix4f>self).c_data[0] = \
                (<ovrMatrix4f>self).c_data[0] * <float>b

        return self

    def __truediv__(ovrMatrix4f a, float b):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = a.c_data[0] * <float>b

        return to_return

    def __itruediv__(self, float b):
        (<ovrMatrix4f>self).c_data[0] = (<ovrMatrix4f>self).c_data[0] / <float>b

        return self

    def transform(self, ovrVector3f v):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).Transform(v.c_data[0])

        return to_return

    def transposed(self):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).Transposed()

        return to_return

    def transpose(self):
        (<ovr_math.Matrix4f>self.c_data[0]).Transpose()

        return self

    def determinant(self):
        cdef float det = (<ovrMatrix4f>self).c_data[0].Determinant()
        return det

    def adjugated(self):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).Adjugated()

        return to_return

    def inverted(self):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).Inverted()

        return to_return

    def invert(self):
        (<ovr_math.Matrix4f>self.c_data[0]).Transpose()

        return self

    def invertedHomogeneousTransform(self):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).InvertedHomogeneousTransform()

        return to_return

    def invertHomogeneousTransform(self):
        (<ovr_math.Matrix4f>self.c_data[0]).InvertHomogeneousTransform()

        return self

    @staticmethod
    def translation(object v):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        if isinstance(v, ovrVector3f):
            (<ovrMatrix4f>to_return).c_data[0] = \
                ovr_math.Matrix4f.Translation((<ovrVector3f>v).c_data[0])
        elif isinstance(v, (list, tuple)):
            (<ovrMatrix4f>to_return).c_data[0] = \
                ovr_math.Matrix4f.Translation(
                    ovrVector3f(<float>v[0], <float>v[1], <float>v[2]).c_data[0])

        return to_return

    def setTranslation(self, ovrVector3f v):
        (<ovr_math.Matrix4f>self.c_data[0]).SetTranslation(v.c_data[0])

        return self

    def getTranslation(self):
        cdef ovrVector3f to_return = ovrVector3f()
        (<ovrVector3f>to_return).c_data[0] = \
            (<ovr_math.Matrix4f>self.c_data[0]).GetTranslation()

        return to_return

    @staticmethod
    def scaling(ovrVector3f v):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = \
            ovr_math.Matrix4f.Scaling((<ovrVector3f>v).c_data[0])

        return to_return

    def distance(self, ovrMatrix4f m2):
        cdef float distance = \
            (<ovr_math.Matrix4f>self.c_data[0]).Distance(m2.c_data[0])

        return distance

    @staticmethod
    def rotationX(float angle=0.0):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = ovr_math.Matrix4f.RotationX(angle)

        return to_return

    @staticmethod
    def rotationY(float angle=0.0):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = ovr_math.Matrix4f.RotationY(angle)

        return to_return

    @staticmethod
    def rotationZ(float angle=0.0):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = ovr_math.Matrix4f.RotationZ(angle)

        return to_return

    @staticmethod
    def lookAt(ovrVector3f eye, ovrVector3f at, ovrVector3f up):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = ovr_math.Matrix4f.LookAtRH(
            eye.c_data[0], at.c_data[0], up.c_data[0])

        return to_return

    @staticmethod
    def perspective(float yfov, float aspect, float znear, float zfar):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = ovr_math.Matrix4f.PerspectiveRH(
            yfov, aspect, znear, zfar)

        return to_return

    @staticmethod
    def ortho2d(float w, float h):
        cdef ovrMatrix4f to_return = ovrMatrix4f()
        (<ovrMatrix4f>to_return).c_data[0] = ovr_math.Matrix4f.Ortho2D(w, h)

        return to_return


# distutils: language=c++
#  =============================================================================
#  OVR Error Code (OVR_ErrorCode.h) Cython Declaration File 
#  =============================================================================
#
#  ovr_errorcode.pxd
#
#  Copyright 2018 Matthew Cutone <cutonem(a)yorku.ca> and Laurie M. Wilcox
#  <lmwilcox(a)yorku.ca>; The Centre For Vision Research, York University, 
#  Toronto, Canada
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in 
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# 
"""This file exposes Oculus Rift(TM) C API types and functions, allowing Cython 
extensions to access them. The declarations in the file are contemporaneous
with version 1.24 (retrieved 04.15.2018) of the Oculus Rift(TM) PC SDK. 

"""
from libc.stdint cimport int32_t

cdef extern from "OVR_ErrorCode.h":
    ctypedef int32_t ovrResult

    ctypedef enum ovrSuccessType:
        ovrSuccess = 0

    ctypedef enum ovrSuccessTypes:
        ovrSuccess_NotVisible = 1000,
        ovrSuccess_BoundaryInvalid = 1001,
        ovrSuccess_DeviceUnavailable = 1002

    ctypedef enum ovrErrorType:
        ovrError_MemoryAllocationFailure = -1000,
        ovrError_InvalidSession = -1002,
        ovrError_Timeout = -1003,
        ovrError_NotInitialized = -1004,
        ovrError_InvalidParameter = -1005,
        ovrError_ServiceError = -1006,
        ovrError_NoHmd = -1007,
        ovrError_Unsupported = -1009,
        ovrError_DeviceUnavailable = -1010,
        ovrError_InvalidHeadsetOrientation = -1011,
        ovrError_ClientSkippedDestroy = -1012,
        ovrError_ClientSkippedShutdown = -1013,
        ovrError_ServiceDeadlockDetected = -1014,
        ovrError_InvalidOperation = -1015,
        ovrError_InsufficientArraySize = -1016,
        ovrError_NoExternalCameraInfo = -1017,
        ovrError_LostTracking = -1018,
        ovrError_ExternalCameraInitializedFailed = -1019,
        ovrError_ExternalCameraCaptureFailed = -1020,
        ovrError_ExternalCameraNameListsBufferSize = -1021,
        ovrError_ExternalCameraNameListsMistmatch = -1022,
        ovrError_ExternalCameraNotCalibrated = -1023,
        ovrError_ExternalCameraNameWrongSize = -1024,
        ovrError_AudioDeviceNotFound = -2001,
        ovrError_AudioComError = -2002,
        ovrError_Initialize = -3000,
        ovrError_LibLoad = -3001,
        ovrError_LibVersion = -3002,
        ovrError_ServiceConnection = -3003,
        ovrError_ServiceVersion = -3004,
        ovrError_IncompatibleOS = -3005,
        ovrError_DisplayInit = -3006,
        ovrError_ServerStart = -3007,
        ovrError_Reinitialization = -3008,
        ovrError_MismatchedAdapters = -3009,
        ovrError_LeakingResources = -3010,
        ovrError_ClientVersion = -3011,
        ovrError_OutOfDateOS = -3012,
        ovrError_OutOfDateGfxDriver = -3013,
        ovrError_IncompatibleGPU = -3014,
        ovrError_NoValidVRDisplaySystem = -3015,
        ovrError_Obsolete = -3016,
        ovrError_DisabledOrDefaultAdapter = -3017,
        ovrError_HybridGraphicsNotSupported = -3018,
        ovrError_DisplayManagerInit = -3019,
        ovrError_TrackerDriverInit = -3020,
        ovrError_LibSignCheck = -3021,
        ovrError_LibPath = -3022,
        ovrError_LibSymbols = -3023,
        ovrError_RemoteSession = -3024,
        ovrError_InitializeVulkan = -3025,
        ovrError_BlacklistedGfxDriver = -3026,
        ovrError_DisplayLost = -6000,
        ovrError_TextureSwapChainFull = -6001,
        ovrError_TextureSwapChainInvalid = -6002,
        ovrError_GraphicsDeviceReset = -6003,
        ovrError_DisplayRemoved = -6004,
        ovrError_ContentProtectionNotAvailable = -6005,
        ovrError_ApplicationInvisible = -6006,
        ovrError_Disallowed = -6007,
        ovrError_DisplayPluggedIncorrectly = -6008,
        ovrError_DisplayLimitReached = -6009,
        ovrError_RuntimeException = -7000,
        ovrError_NoCalibration = -9000,
        ovrError_OldVersion = -9001,
        ovrError_MisformattedBlock = -9002

    ctypedef struct ovrErrorInfo:
        ovrResult Result
        char[512] ErrorString

cdef inline int OVR_SUCCESS(ovrResult result):
    return result >= ovrSuccessType.ovrSuccess

cdef inline int OVR_UNQUALIFIED_SUCCESS(ovrResult result):
    return result == ovrSuccessType.ovrSuccess

cdef inline int OVR_FAILURE(ovrResult result):
    return not OVR_SUCCESS(result)


# distutils: language=c++
#  =============================================================================
#  OVR Misc. Utilities (OVR_CAPI_Util.h) Cython Declaration File 
#  =============================================================================
#
#  ovr_capi_util.pxd
#
#  Copyright 2018 Matthew Cutone <cutonem(a)yorku.ca> and Laurie M. Wilcox
#  <lmwilcox(a)yorku.ca>; The Centre For Vision Research, York University, 
#  Toronto, Canada
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in 
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# 
"""This file exposes Oculus Rift(TM) C API types and functions, allowing Cython 
extensions to access them. The declarations in the file are contemporaneous
with version 1.24 (retrieved 04.15.2018) of the Oculus Rift(TM) PC SDK. 

"""
from ovr_capi cimport ovrBool, ovrResult, ovrSession
from ovr_capi cimport ovrMatrix4f, ovrFovPort, ovrVector2f, ovrVector3f, ovrPosef
from ovr_capi cimport ovrTimewarpProjectionDesc

cdef extern from "OVR_CAPI_Util.h":
    ctypedef enum ovrProjectionModifier:
        ovrProjection_None = 0x00,
        ovrProjection_LeftHanded = 0x01,
        ovrProjection_FarLessThanNear = 0x02,
        ovrProjection_FarClipAtInfinity = 0x04,
        ovrProjection_ClipRangeOpenGL = 0x08

    ctypedef struct ovrDetectResult:
        ovrBool IsOculusServiceRunning
        ovrBool IsOculusHMDConnected

    ctypedef enum ovrHapticsGenMode:
        ovrHapticsGenMode_PointSample,
        ovrHapticsGenMode_Count

    ctypedef struct ovrAudioChannelData:
        const float* Samples
        int SamplesCount
        int Frequency

    ctypedef struct ovrHapticsClip:
        const void* Samples
        int SamplesCount

    cdef ovrDetectResult ovr_Detect(int timeoutMilliseconds)
    cdef ovrMatrix4f ovrMatrix4f_Projection(ovrFovPort fov, float znear, float zfar, unsigned int projectionModFlags)
    cdef ovrTimewarpProjectionDesc ovrTimewarpProjectionDesc_FromProjection(ovrMatrix4f projection, unsigned int projectionModFlags)
    cdef ovrMatrix4f ovrMatrix4f_OrthoSubProjection(ovrMatrix4f projection, ovrVector2f orthoScale, float orthoDistance, float HmdToEyeOffsetX)
    cdef void ovr_CalcEyePoses(ovrPosef headPose, const ovrVector3f hmdToEyeOffset[2], ovrPosef outEyePoses[2])
    cdef void ovr_CalcEyePoses2(ovrPosef headPose, const ovrPosef HmdToEyePose[2], ovrPosef outEyePoses[2])
    cdef void ovr_GetEyePoses(ovrSession session, long long frameIndex, ovrBool latencyMarker, const ovrVector3f hmdToEyeOffset[2], ovrPosef outEyePoses[2], double* outSensorSampleTime)
    cdef void ovr_GetEyePoses2(ovrSession session, long long frameIndex, ovrBool latencyMarker, const ovrPosef HmdToEyePose[2], ovrPosef outEyePoses[2], double* outSensorSampleTime)
    cdef void ovrPosef_FlipHandedness(const ovrPosef* inPose, ovrPosef* outPose)
    cdef ovrResult ovr_ReadWavFromBuffer(ovrAudioChannelData* outAudioChannel, const void* inputData, int dataSizeInBytes, int stereoChannelToUse)
    cdef ovrResult ovr_GenHapticsFromAudioData(ovrHapticsClip* outHapticsClip, const ovrAudioChannelData* audioChannel, ovrHapticsGenMode genMode)
    cdef void ovr_ReleaseAudioChannelData(ovrAudioChannelData* audioChannel)
    cdef void ovr_ReleaseHapticsClip(ovrHapticsClip* hapticsClip)
